from __future__ import annotations

import datetime as dt
import os
from pathlib import Path as P
import uuid
from pathlib import Path
from typing import Annotated, List, Optional

import typer

from iac.cli.apps.exec import app as exec_app
from iac.runbook.discovery import (
    parse_book_dirs,
    parse_module_dirs,
    resolve_book,
    split_module_function,
    split_book_step,
    book_id_for_path,
)
from iac.runbook.loader import load_book, parse_vars_kv, resolve_inputs
from iac.runbook.runner import RunOptions, run_book_sequential
from iac.runbook.model import Book, BookMeta, Step
from iac.cli.formatting import ensure_output_format, emit_structured, build_table

app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Execute commands, books, steps, and functions",
)

# Mount command runner under `iac run cmd`
app.add_typer(exec_app, name="", help="")


def _book_dirs() -> List[Path]:
    env = os.environ.get("IAC_BOOK_DIRS")
    return parse_book_dirs(env)


def _module_dirs() -> List[Path]:
    env = os.environ.get("IAC_MODULE_DIRS")
    return parse_module_dirs(env)


def _opts(
    quiet: bool,
    stdin_file: Optional[str],
    stdout_file: Optional[str],
    stderr_file: Optional[str],
    file_mode: str,
    artifacts_dir: Optional[str],
    capture: bool,
    print_captured: bool,
    pty: bool,
    killpg: bool,
    forward_signals: bool,
    timeout: Optional[float],
    retries: Optional[int],
    retry_delay: Optional[float],
    verbose: bool,
    run_id: str,
) -> RunOptions:
    return RunOptions(
        quiet=quiet,
        stdin_file=stdin_file,
        stdout_file=stdout_file,
        stderr_file=stderr_file,
        file_mode=file_mode,
        artifacts_dir=artifacts_dir,
        capture=capture,
        print_captured=print_captured,
        pty=pty,
        killpg=killpg,
        forward_signals=forward_signals,
        timeout_override=timeout,
        retries_override=retries,
        retry_delay_override=retry_delay,
        verbose=verbose,
        run_id=run_id,
    )


def _select_steps(
    steps: List[Step],
    *,
    from_step: Optional[str],
    to_step: Optional[str],
    only_steps: List[str],
    skip_steps: List[str],
    tags: List[str],
) -> List[Step]:
    by_name = {s.name: s for s in steps}
    selected = list(steps)

    if only_steps and (from_step or to_step):
        raise typer.BadParameter(
            "--only-step cannot be combined with --from-step/--to-step"
        )

    if only_steps:
        missing = [n for n in only_steps if n not in by_name]
        if missing:
            raise typer.BadParameter(
                f"Unknown --only-step value(s): {', '.join(missing)}"
            )
        only_set = set(only_steps)
        selected = [s for s in selected if s.name in only_set]
    else:
        start_idx = 0
        end_idx = len(selected) - 1
        if from_step:
            if from_step not in by_name:
                raise typer.BadParameter(f"Unknown --from-step: {from_step}")
            start_idx = next(i for i, s in enumerate(selected) if s.name == from_step)
        if to_step:
            if to_step not in by_name:
                raise typer.BadParameter(f"Unknown --to-step: {to_step}")
            end_idx = next(i for i, s in enumerate(selected) if s.name == to_step)
        if start_idx > end_idx:
            raise typer.BadParameter(
                "--from-step must come before --to-step in book order"
            )
        selected = selected[start_idx : end_idx + 1]

    if tags:
        tag_set = set(tags)
        selected = [s for s in selected if any(t in tag_set for t in s.tags)]

    if skip_steps:
        unknown_skip = [n for n in skip_steps if n not in by_name]
        if unknown_skip:
            raise typer.BadParameter(
                f"Unknown --skip-step value(s): {', '.join(unknown_skip)}"
            )
        skip_set = set(skip_steps)
        selected = [s for s in selected if s.name not in skip_set]

    if not selected:
        raise typer.BadParameter("Step selection produced an empty run set")
    return selected


def _results_summary(results: list) -> dict:
    out = {
        "total": len(results),
        "success": 0,
        "skipped": 0,
        "failed": 0,
        "failed_nonblocking": 0,
    }
    for r in results:
        out[r.status] = out.get(r.status, 0) + 1
    return out


def _compact_command(command: str, *, max_chars: int = 96) -> str:
    s = (command or "").strip()
    if not s:
        return ""
    lines = [ln.strip() for ln in s.splitlines() if ln.strip()]
    if not lines:
        return ""
    head = lines[0]
    if len(lines) > 1:
        head = f"{head} ... (+{len(lines) - 1} lines)"
    if len(head) > max_chars:
        return head[: max_chars - 3] + "..."
    return head


def _resolve_report_path(
    report: Optional[str], report_dir: Optional[str], run_id: str, fmt: str
) -> Optional[str]:
    if report:
        return str(P(report).expanduser().resolve())
    if report_dir:
        base = P(report_dir).expanduser().resolve()
        base.mkdir(parents=True, exist_ok=True)
        ext = "yaml" if fmt == "yaml" else "json"
        return str(base / f"{run_id}.{ext}")
    return None


def _write_report(path: Optional[str], payload: dict, fmt: str) -> Optional[str]:
    if not path:
        return None
    p = P(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    if fmt == "yaml":
        import yaml

        p.write_text(
            yaml.safe_dump(payload, sort_keys=False, default_flow_style=False),
            encoding="utf-8",
        )
    else:
        import json

        p.write_text(json.dumps(payload, indent=2), encoding="utf-8")
    return str(p)


def _run_table(results: List[object], *, verbose: bool):
    if verbose:
        headers = [
            "STEP",
            "STATUS",
            "ATTEMPTS",
            "EXIT",
            "DURATION_MS",
            "CWD",
            "STDOUT",
            "STDERR",
            "COMMAND",
            "MESSAGE",
        ]
        rows = [
            [
                r.step,
                r.status,
                f"{r.attempts}/{r.max_attempts}",
                r.exit_code,
                r.duration_ms,
                r.cwd,
                r.stdout_file,
                r.stderr_file,
                _compact_command(r.command),
                r.message,
            ]
            for r in results
        ]
        return build_table(headers, rows)

    headers = ["STEP", "STATUS", "ATTEMPTS", "EXIT", "DURATION_MS", "MESSAGE"]
    rows = [
        [
            r.step,
            r.status,
            f"{r.attempts}/{r.max_attempts}",
            r.exit_code,
            r.duration_ms,
            r.message,
        ]
        for r in results
    ]
    return build_table(headers, rows)


@app.command("book")
def run_book(
    book: Annotated[str, typer.Argument(help="Book name or path (.yaml/.yml).")],
    input: Annotated[
        List[str],
        typer.Option(
            "--input",
            help="Provide runbook inputs (repeatable): --input key=value",
            show_default=False,
        ),
    ] = [],
    output: Annotated[
        str,
        typer.Option(
            "--output", help="Output format: table|json|yaml", show_default=True
        ),
    ] = "table",
    report: Annotated[
        Optional[str],
        typer.Option("--report", help="Write run report to this file path"),
    ] = None,
    report_dir: Annotated[
        Optional[str],
        typer.Option("--report-dir", help="Write run report under this directory"),
    ] = None,
    report_format: Annotated[
        str,
        typer.Option(
            "--report-format", help="Report format: json|yaml", show_default=True
        ),
    ] = "json",
    # runner options
    quiet: Annotated[
        bool, typer.Option("-q", "--quiet", help="Do not stream output")
    ] = False,
    stdin_file: Annotated[
        Optional[str],
        typer.Option("--stdin-file", help="Provide stdin bytes from file"),
    ] = None,
    stdout_file: Annotated[
        Optional[str], typer.Option("--stdout-file", help="Append stdout bytes to file")
    ] = None,
    stderr_file: Annotated[
        Optional[str], typer.Option("--stderr-file", help="Append stderr bytes to file")
    ] = None,
    file_mode: Annotated[
        str,
        typer.Option(
            "--file-mode", help="File write mode: append|truncate", show_default=True
        ),
    ] = "append",
    artifacts_dir: Annotated[
        Optional[str],
        typer.Option(
            "--artifacts-dir",
            help="Auto-write per-step stdout/stderr artifacts in this directory",
        ),
    ] = None,
    capture: Annotated[
        bool, typer.Option("--capture/--no-capture", help="Capture output in memory")
    ] = False,
    print_captured: Annotated[
        bool,
        typer.Option(
            "--print-captured/--no-print-captured",
            help="If --capture, print captured at end",
        ),
    ] = False,
    pty: Annotated[bool, typer.Option("--pty", help="Run under a PTY")] = False,
    timeout: Annotated[
        Optional[float],
        typer.Option(
            "--timeout", help="Override timeout seconds (applies to each step)"
        ),
    ] = None,
    retries: Annotated[
        Optional[int],
        typer.Option("--retries", help="Override retries for each step (>=0)"),
    ] = None,
    retry_delay: Annotated[
        Optional[float],
        typer.Option(
            "--retry-delay", help="Override retry delay seconds for each step (>=0)"
        ),
    ] = None,
    from_step: Annotated[
        Optional[str],
        typer.Option("--from-step", help="Start at this step name (inclusive)"),
    ] = None,
    to_step: Annotated[
        Optional[str],
        typer.Option("--to-step", help="Stop at this step name (inclusive)"),
    ] = None,
    only_step: Annotated[
        List[str],
        typer.Option("--only-step", help="Run only these step names (repeatable)"),
    ] = [],
    skip_step: Annotated[
        List[str],
        typer.Option("--skip-step", help="Skip these step names (repeatable)"),
    ] = [],
    tag: Annotated[
        List[str],
        typer.Option(
            "--tag", help="Run only steps containing any provided tag (repeatable)"
        ),
    ] = [],
    killpg: Annotated[
        bool, typer.Option("--killpg/--no-killpg", help="Use process-group termination")
    ] = True,
    forward_signals: Annotated[
        bool,
        typer.Option(
            "--forward-signals/--no-forward-signals", help="Forward signals to child"
        ),
    ] = True,
    verbose: Annotated[
        bool, typer.Option("-v", "--verbose", help="Verbose process metadata")
    ] = False,
) -> None:
    """Run all steps in a book sequentially."""
    dirs = _book_dirs()
    module_dirs = _module_dirs()
    output_fmt = ensure_output_format(output)
    if report_format not in ("json", "yaml"):
        raise typer.BadParameter("Invalid --report-format. Use json or yaml.")
    if file_mode not in ("append", "truncate"):
        raise typer.BadParameter("Invalid --file-mode. Use append or truncate.")
    if retries is not None and retries < 0:
        raise typer.BadParameter("--retries must be >= 0")
    if retry_delay is not None and retry_delay < 0:
        raise typer.BadParameter("--retry-delay must be >= 0")

    run_id = uuid.uuid4().hex[:12]
    report_path = _resolve_report_path(report, report_dir, run_id, report_format)
    bp = resolve_book(book, dirs)
    b = load_book(bp)
    vars_map = dict(resolve_inputs(b.inputs, input_cli=input))
    opts = _opts(
        quiet,
        stdin_file,
        stdout_file,
        stderr_file,
        file_mode,
        artifacts_dir,
        capture,
        print_captured,
        pty,
        killpg,
        forward_signals,
        timeout,
        retries,
        retry_delay,
        verbose,
        run_id,
    )
    selected_steps = _select_steps(
        b.steps,
        from_step=from_step,
        to_step=to_step,
        only_steps=only_step,
        skip_steps=skip_step,
        tags=tag,
    )
    run_book_obj = Book(
        meta=b.meta,
        vars=b.vars,
        inputs=b.inputs,
        requires=b.requires,
        outputs=b.outputs,
        hooks=b.hooks,
        sensitive=b.sensitive,
        steps=selected_steps,
    )

    started_at = dt.datetime.now(dt.timezone.utc).isoformat()
    results, final_outputs = run_book_sequential(
        run_book_obj,
        book_path=bp,
        vars_map=vars_map,
        opts=opts,
        report_path=report_path,
        module_dirs=module_dirs,
    )
    ended_at = dt.datetime.now(dt.timezone.utc).isoformat()

    exit_code = 0
    for r in results:
        if r.status == "failed":
            exit_code = r.exit_code or 1
            break

    payload = {
        "report_version": 1,
        "run_id": run_id,
        "kind": "book",
        "book": str(bp),
        "started_at": started_at,
        "ended_at": ended_at,
        "inputs": vars_map,
        "outputs": final_outputs,
        "selection": {
            "from_step": from_step or "",
            "to_step": to_step or "",
            "only_steps": only_step,
            "skip_steps": skip_step,
            "tags": tag,
        },
        "summary": _results_summary(results),
        "results": [r.__dict__ for r in results],
        "exit_code": exit_code,
    }
    written_report = _write_report(report_path, payload, report_format)
    if written_report:
        payload["report_path"] = written_report

    if output_fmt in ("json", "yaml"):
        emit_structured(payload, output_fmt)
        raise typer.Exit(exit_code)

    table = _run_table(results, verbose=verbose)
    typer.echo(f"Run Book: {bp}")
    typer.echo(table.get_string())
    summary = _results_summary(results)
    typer.echo(
        "Summary: "
        f"total={summary['total']} success={summary['success']} skipped={summary['skipped']} "
        f"failed_nonblocking={summary['failed_nonblocking']} failed={summary['failed']}"
    )
    failed = next((r for r in results if r.status == "failed"), None)
    if failed is not None:
        book_ref = book_id_for_path(bp)
        typer.echo(
            f"Failure: step={failed.step} exit={failed.exit_code} attempts={failed.attempts}/{failed.max_attempts}"
        )
        if failed.message:
            typer.echo(f"Reason: {failed.message}")
        typer.echo(f"Rerun Hint: iac run step {book_ref}@{failed.step}")
    if written_report:
        typer.echo(f"Report: {written_report}")
    typer.echo(f"Exit Code: {exit_code}")
    raise typer.Exit(exit_code)


@app.command("step")
def run_step(
    ref: Annotated[
        str,
        typer.Argument(
            help="Step reference 'book@step'. Book resolved from IAC_BOOK_DIRS or path."
        ),
    ],
    input: Annotated[
        List[str],
        typer.Option(
            "--input",
            help="Provide runbook inputs (repeatable): --input key=value",
            show_default=False,
        ),
    ] = [],
    output: Annotated[
        str,
        typer.Option(
            "--output", help="Output format: table|json|yaml", show_default=True
        ),
    ] = "table",
    report: Annotated[
        Optional[str],
        typer.Option("--report", help="Write run report to this file path"),
    ] = None,
    report_dir: Annotated[
        Optional[str],
        typer.Option("--report-dir", help="Write run report under this directory"),
    ] = None,
    report_format: Annotated[
        str,
        typer.Option(
            "--report-format", help="Report format: json|yaml", show_default=True
        ),
    ] = "json",
    # runner options
    quiet: Annotated[
        bool, typer.Option("-q", "--quiet", help="Do not stream output")
    ] = False,
    stdin_file: Annotated[
        Optional[str],
        typer.Option("--stdin-file", help="Provide stdin bytes from file"),
    ] = None,
    stdout_file: Annotated[
        Optional[str], typer.Option("--stdout-file", help="Append stdout bytes to file")
    ] = None,
    stderr_file: Annotated[
        Optional[str], typer.Option("--stderr-file", help="Append stderr bytes to file")
    ] = None,
    file_mode: Annotated[
        str,
        typer.Option(
            "--file-mode", help="File write mode: append|truncate", show_default=True
        ),
    ] = "append",
    artifacts_dir: Annotated[
        Optional[str],
        typer.Option(
            "--artifacts-dir",
            help="Auto-write per-step stdout/stderr artifacts in this directory",
        ),
    ] = None,
    capture: Annotated[
        bool, typer.Option("--capture/--no-capture", help="Capture output in memory")
    ] = False,
    print_captured: Annotated[
        bool,
        typer.Option(
            "--print-captured/--no-print-captured",
            help="If --capture, print captured at end",
        ),
    ] = False,
    pty: Annotated[bool, typer.Option("--pty", help="Run under a PTY")] = False,
    timeout: Annotated[
        Optional[float], typer.Option("--timeout", help="Override timeout seconds")
    ] = None,
    retries: Annotated[
        Optional[int],
        typer.Option("--retries", help="Override retries for this step (>=0)"),
    ] = None,
    retry_delay: Annotated[
        Optional[float],
        typer.Option("--retry-delay", help="Override retry delay seconds (>=0)"),
    ] = None,
    killpg: Annotated[
        bool, typer.Option("--killpg/--no-killpg", help="Use process-group termination")
    ] = True,
    forward_signals: Annotated[
        bool,
        typer.Option(
            "--forward-signals/--no-forward-signals", help="Forward signals to child"
        ),
    ] = True,
    verbose: Annotated[
        bool, typer.Option("-v", "--verbose", help="Verbose process metadata")
    ] = False,
) -> None:
    """Run a single step from a book (book@step)."""
    dirs = _book_dirs()
    module_dirs = _module_dirs()
    output_fmt = ensure_output_format(output)
    if report_format not in ("json", "yaml"):
        raise typer.BadParameter("Invalid --report-format. Use json or yaml.")
    if file_mode not in ("append", "truncate"):
        raise typer.BadParameter("Invalid --file-mode. Use append or truncate.")
    if retries is not None and retries < 0:
        raise typer.BadParameter("--retries must be >= 0")
    if retry_delay is not None and retry_delay < 0:
        raise typer.BadParameter("--retry-delay must be >= 0")

    run_id = uuid.uuid4().hex[:12]
    report_path = _resolve_report_path(report, report_dir, run_id, report_format)
    book_ref, step_name = split_book_step(ref)
    bp = resolve_book(book_ref, dirs)
    b = load_book(bp)

    step = next((s for s in b.steps if s.name == step_name), None)
    if step is None:
        typer.echo(f"Unknown step '{step_name}' in book {bp}", err=True)
        raise typer.Exit(2)

    vars_map = dict(resolve_inputs(b.inputs, input_cli=input))
    opts = _opts(
        quiet,
        stdin_file,
        stdout_file,
        stderr_file,
        file_mode,
        artifacts_dir,
        capture,
        print_captured,
        pty,
        killpg,
        forward_signals,
        timeout,
        retries,
        retry_delay,
        verbose,
        run_id,
    )

    started_at = dt.datetime.now(dt.timezone.utc).isoformat()
    run_book_obj = b.__class__(
        meta=b.meta,
        vars=b.vars,
        inputs=b.inputs,
        requires=b.requires,
        outputs=b.outputs,
        hooks=b.hooks,
        sensitive=b.sensitive,
        steps=[step],
    )
    results, final_outputs = run_book_sequential(
        run_book_obj,
        book_path=bp,
        vars_map=vars_map,
        opts=opts,
        report_path=report_path,
        module_dirs=module_dirs,
    )
    ended_at = dt.datetime.now(dt.timezone.utc).isoformat()
    res = results[0]

    exit_code = (
        0
        if res.status in ("success", "skipped", "failed_nonblocking")
        else (res.exit_code or 1)
    )

    payload = {
        "report_version": 1,
        "run_id": run_id,
        "kind": "step",
        "book": str(bp),
        "ref": ref,
        "started_at": started_at,
        "ended_at": ended_at,
        "result": res.__dict__,
        "outputs": final_outputs,
        "summary": _results_summary(results),
        "exit_code": exit_code,
    }
    written_report = _write_report(report_path, payload, report_format)
    if written_report:
        payload["report_path"] = written_report

    if output_fmt in ("json", "yaml"):
        emit_structured(payload, output_fmt)
        raise typer.Exit(exit_code)

    table = _run_table([res], verbose=verbose)
    typer.echo(f"Run Step: {bp.name}@{res.step}")
    typer.echo(table.get_string())
    if res.status == "failed":
        book_ref = book_id_for_path(bp)
        typer.echo(
            f"Failure: step={res.step} exit={res.exit_code} attempts={res.attempts}/{res.max_attempts}"
        )
        if res.message:
            typer.echo(f"Reason: {res.message}")
        typer.echo(f"Rerun Hint: iac run step {book_ref}@{res.step}")
    if written_report:
        typer.echo(f"Report: {written_report}")
    typer.echo(f"Exit Code: {exit_code}")
    raise typer.Exit(exit_code)


@app.command("fn")
def run_fn(
    ref: Annotated[
        str,
        typer.Argument(
            help="Function reference 'module@function'. Module resolved from IAC_MODULE_DIRS."
        ),
    ],
    var: Annotated[
        List[str],
        typer.Option(
            "--var", help="Override vars (repeatable): --var k=v", show_default=False
        ),
    ] = [],
    output: Annotated[
        str,
        typer.Option(
            "--output", help="Output format: table|json|yaml", show_default=True
        ),
    ] = "table",
    report: Annotated[
        Optional[str],
        typer.Option("--report", help="Write run report to this file path"),
    ] = None,
    report_dir: Annotated[
        Optional[str],
        typer.Option("--report-dir", help="Write run report under this directory"),
    ] = None,
    report_format: Annotated[
        str,
        typer.Option(
            "--report-format", help="Report format: json|yaml", show_default=True
        ),
    ] = "json",
    # runner options
    quiet: Annotated[
        bool, typer.Option("-q", "--quiet", help="Do not stream output")
    ] = False,
    stdin_file: Annotated[
        Optional[str],
        typer.Option("--stdin-file", help="Provide stdin bytes from file"),
    ] = None,
    stdout_file: Annotated[
        Optional[str], typer.Option("--stdout-file", help="Append stdout bytes to file")
    ] = None,
    stderr_file: Annotated[
        Optional[str], typer.Option("--stderr-file", help="Append stderr bytes to file")
    ] = None,
    file_mode: Annotated[
        str,
        typer.Option(
            "--file-mode", help="File write mode: append|truncate", show_default=True
        ),
    ] = "append",
    artifacts_dir: Annotated[
        Optional[str],
        typer.Option(
            "--artifacts-dir",
            help="Auto-write per-step stdout/stderr artifacts in this directory",
        ),
    ] = None,
    capture: Annotated[
        bool, typer.Option("--capture/--no-capture", help="Capture output in memory")
    ] = False,
    print_captured: Annotated[
        bool,
        typer.Option(
            "--print-captured/--no-print-captured",
            help="If --capture, print captured at end",
        ),
    ] = False,
    pty: Annotated[bool, typer.Option("--pty", help="Run under a PTY")] = False,
    timeout: Annotated[
        Optional[float], typer.Option("--timeout", help="Override timeout seconds")
    ] = None,
    retries: Annotated[
        Optional[int],
        typer.Option("--retries", help="Override retries for this function (>=0)"),
    ] = None,
    retry_delay: Annotated[
        Optional[float],
        typer.Option("--retry-delay", help="Override retry delay seconds (>=0)"),
    ] = None,
    killpg: Annotated[
        bool, typer.Option("--killpg/--no-killpg", help="Use process-group termination")
    ] = True,
    forward_signals: Annotated[
        bool,
        typer.Option(
            "--forward-signals/--no-forward-signals", help="Forward signals to child"
        ),
    ] = True,
    verbose: Annotated[
        bool, typer.Option("-v", "--verbose", help="Verbose process metadata")
    ] = False,
) -> None:
    """Run a Python module function directly (module@function)."""
    try:
        _module_name, fn_name = split_module_function(ref)
    except ValueError as e:
        raise typer.BadParameter(str(e)) from e

    module_dirs = _module_dirs()
    output_fmt = ensure_output_format(output)
    if report_format not in ("json", "yaml"):
        raise typer.BadParameter("Invalid --report-format. Use json or yaml.")
    if file_mode not in ("append", "truncate"):
        raise typer.BadParameter("Invalid --file-mode. Use append or truncate.")
    if retries is not None and retries < 0:
        raise typer.BadParameter("--retries must be >= 0")
    if retry_delay is not None and retry_delay < 0:
        raise typer.BadParameter("--retry-delay must be >= 0")

    run_id = uuid.uuid4().hex[:12]
    report_path = _resolve_report_path(report, report_dir, run_id, report_format)
    vars_map = parse_vars_kv(var)
    opts = _opts(
        quiet,
        stdin_file,
        stdout_file,
        stderr_file,
        file_mode,
        artifacts_dir,
        capture,
        print_captured,
        pty,
        killpg,
        forward_signals,
        timeout,
        retries,
        retry_delay,
        verbose,
        run_id,
    )

    synthetic_step = Step(
        name=fn_name,
        mode="module",
        module=ref,
        on_fail="halt",
    )
    synthetic_book = Book(
        meta=BookMeta(name="__fn__"),
        vars={},
        inputs={},
        requires=[],
        outputs={},
        hooks={},
        sensitive=[],
        steps=[synthetic_step],
    )
    synthetic_book_path = P.cwd() / ".iac" / "__fn__.yaml"

    started_at = dt.datetime.now(dt.timezone.utc).isoformat()
    results, final_outputs = run_book_sequential(
        synthetic_book,
        book_path=synthetic_book_path,
        vars_map=vars_map,
        opts=opts,
        report_path=report_path,
        module_dirs=module_dirs,
    )
    ended_at = dt.datetime.now(dt.timezone.utc).isoformat()
    res = results[0]

    exit_code = (
        0
        if res.status in ("success", "skipped", "failed_nonblocking")
        else (res.exit_code or 1)
    )

    payload = {
        "report_version": 1,
        "run_id": run_id,
        "kind": "fn",
        "ref": ref,
        "started_at": started_at,
        "ended_at": ended_at,
        "vars": vars_map,
        "outputs": final_outputs,
        "result": res.__dict__,
        "summary": _results_summary(results),
        "exit_code": exit_code,
    }
    written_report = _write_report(report_path, payload, report_format)
    if written_report:
        payload["report_path"] = written_report

    if output_fmt in ("json", "yaml"):
        emit_structured(payload, output_fmt)
        raise typer.Exit(exit_code)

    table = _run_table([res], verbose=verbose)
    typer.echo(f"Run Function: {ref}")
    typer.echo(table.get_string())
    if res.status == "failed":
        typer.echo(
            f"Failure: fn={ref} exit={res.exit_code} attempts={res.attempts}/{res.max_attempts}"
        )
        if res.message:
            typer.echo(f"Reason: {res.message}")
        typer.echo(f"Rerun Hint: iac run fn {ref}")
    if written_report:
        typer.echo(f"Report: {written_report}")
    typer.echo(f"Exit Code: {exit_code}")
    raise typer.Exit(exit_code)



from __future__ import annotations

from iac.runbook.cel_eval import CelEngine, _to_python
from iac.runbook.expr import default_ctx
from iac.runbook.model import Book, BookMeta
from iac.runbook.sensitive import redact_vars_map, sensitive_values
from iac.runbook.template import TemplateEngine, _filter_default, _filter_replace


def test_cel_engine_empty_and_type_conversions() -> None:
    cel = CelEngine.default()
    assert cel.eval("", {"inputs": {}, "steps": {}, "run": {}}) is None
    assert isinstance(cel.eval("1.5", {}), float)
    assert cel.eval("[1,2,3]", {}) == [1, 2, 3]
    assert cel.eval('{"a": 1}', {}) == {"a": 1}


def test_cel_to_python_fallback_exception_path() -> None:
    cyc = []
    cyc.append(cyc)
    out = _to_python(cyc)
    assert out is cyc


def test_template_and_expr_misc_paths() -> None:
    t = TemplateEngine.default()
    assert t.render_str("${HOME}", vars_map={}, env_map={"HOME": "/tmp"}) == "/tmp"
    assert t.render_obj(None, vars_map={}, env_map={}) is None
    assert t.render_obj(
        ["${X}", {"k": "${Y}"}], vars_map={}, env_map={"X": "a", "Y": "b"}
    ) == ["a", {"k": "b"}]
    assert _filter_default(None, "x") == "x"
    assert _filter_replace(None, "a", "b") == ""
    ctx = default_ctx()
    assert "os" in ctx


def test_sensitive_helpers_non_empty_keys() -> None:
    book = Book(
        meta=BookMeta(name="x"),
        vars={},
        inputs={},
        requires=[],
        outputs={},
        hooks={},
        sensitive=["secret"],
        steps=[],
    )
    vals = sensitive_values(book, vars_map={"secret": "top"})
    assert vals == ["top"]
    redacted = redact_vars_map(book, {"secret": "top", "ok": "1"})
    assert redacted["secret"] == "********"
    assert redacted["ok"] == "1"



from __future__ import annotations

import json
import runpy
from pathlib import Path

import pytest

from iac.runbook import module_exec
from iac.runbook.module_contract import load_validated_callable
from iac.runbook.module_types import StepContext, StepResult


def _write_module(path: Path, src: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(src, encoding="utf-8")


def test_step_context_and_result_helpers() -> None:
    ctx = StepContext(
        vars={"a": 1},
        env={},
        book_path="book.yaml",
        step_name="s",
        run_id="r",
        cwd=".",
    )
    assert ctx.var("a") == 1
    assert ctx.var("missing", "d") == "d"
    assert ctx.require_var("a") == 1
    with pytest.raises(KeyError):
        ctx.require_var("missing")

    r = StepResult(message="ok", outputs={"x": 1})
    assert r.message == "ok"
    assert r.outputs["x"] == 1


def test_module_exec_main_error_paths(
    tmp_path: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    rc = module_exec.main([])
    assert rc == 2
    assert "missing module ref" in capsys.readouterr().err

    rc = module_exec.main(["--ref", "ops@run"])
    assert rc == 2
    assert "missing step name" in capsys.readouterr().err

    rc = module_exec.main(
        ["--ref", "ops@run", "--step-name", "s", "--vars-json", "[1]"]
    )
    assert rc == 2
    assert "invalid --vars-json" in capsys.readouterr().err
    # no-op path write helper branch
    module_exec._write_result(None, {"message": "x"})


def test_module_exec_main_success_and_result_shapes(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    modules = tmp_path / "modules"
    _write_module(
        modules / "ops.py",
        "\n".join(
            [
                "from iac.runbook.module_types import StepContext, StepResult",
                "def returns_result(ctx: StepContext) -> StepResult:",
                "    '''return a valid result'''",
                "    return StepResult(message='ok', outputs={'a': 1})",
                "def returns_empty(ctx: StepContext) -> StepResult:",
                "    '''return empty result'''",
                "    return StepResult()",
                "def returns_dict(ctx: StepContext) -> StepResult:",
                "    '''invalid runtime return'''",
                "    return {'message': 'dict'}  # type: ignore[return-value]",
                "def boom(ctx: StepContext) -> StepResult:",
                "    '''raise runtime error'''",
                "    raise RuntimeError('boom')",
                "",
            ]
        ),
    )
    monkeypatch.setenv("IAC_MODULE_DIRS", str(modules))

    def run_one(fn: str) -> dict:
        result_file = tmp_path / f"{fn}.json"
        rc = module_exec.main(
            [
                "--ref",
                f"ops@{fn}",
                "--step-name",
                "s1",
                "--book-path",
                str(tmp_path / "book.yaml"),
                "--run-id",
                "rid",
                "--vars-json",
                '{"x":"1"}',
                "--result-file",
                str(result_file),
            ]
        )
        assert rc == 0
        return json.loads(result_file.read_text(encoding="utf-8"))

    assert run_one("returns_result") == {"message": "ok", "outputs": {"a": 1}}
    assert run_one("returns_empty") == {"message": "", "outputs": {}}

    rc = module_exec.main(["--ref", "ops@returns_dict", "--step-name", "s1"])
    assert rc == 1
    rc = module_exec.main(["--ref", "ops@boom", "--step-name", "s1"])
    assert rc == 1


def test_module_contract_load_callable_validation(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    modules = tmp_path / "modules"
    _write_module(
        modules / "bad.py",
        "\n".join(
            [
                "value = 1",
                "from iac.runbook.module_types import StepContext, StepResult",
                "def bad_sig(ctx: StepContext, extra) -> StepResult:",
                "    '''bad signature'''",
                "    return StepResult()",
                "def camelCase(ctx: StepContext) -> StepResult:",
                "    '''bad name'''",
                "    return StepResult()",
                "def no_doc(ctx: StepContext) -> StepResult:",
                "    return StepResult()",
                "def bad_param(ctx: str) -> StepResult:",
                "    '''bad param'''",
                "    return StepResult()",
                "def bad_return(ctx: StepContext) -> dict:",
                "    '''bad return'''",
                "    return {}",
                "",
            ]
        ),
    )
    monkeypatch.setenv("IAC_MODULE_DIRS", str(modules))

    with pytest.raises(AttributeError):
        load_validated_callable("bad", "missing")
    with pytest.raises(TypeError):
        load_validated_callable("bad", "bad_sig")
    with pytest.raises(ValueError):
        load_validated_callable("bad", "camelCase")
    with pytest.raises(ValueError):
        load_validated_callable("bad", "no_doc")
    with pytest.raises(TypeError):
        load_validated_callable("bad", "bad_param")
    with pytest.raises(TypeError):
        load_validated_callable("bad", "bad_return")


def test_module_can_import_step_types_from_iac(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    modules = tmp_path / "modules"
    _write_module(
        modules / "imports_iac.py",
        "\n".join(
            [
                "from iac import StepContext, StepResult",
                "def run(ctx: StepContext) -> StepResult:",
                "    '''imports through iac root'''",
                "    return StepResult(message='ok')",
                "",
            ]
        ),
    )
    monkeypatch.setenv("IAC_MODULE_DIRS", str(modules))
    fn = load_validated_callable("imports_iac", "run")
    assert callable(fn)


def test_module_exec_main_module_path_and_package_init_coverage(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Execute package __init__ file directly so coverage records iac/__init__.py lines.
    init_ns = runpy.run_path("iac/__init__.py")
    assert "StepContext" in init_ns
    assert "StepResult" in init_ns

    # Execute module_exec as __main__ and catch the SystemExit.
    monkeypatch.setenv("IAC_MODULE_REF", "")
    monkeypatch.setenv("IAC_MODULE_STEP_NAME", "")
    with pytest.raises(SystemExit):
        runpy.run_module("iac.runbook.module_exec", run_name="__main__")



from __future__ import annotations

import datetime as dt
import json
import os
import re
import shlex
import signal
import sys
import tempfile
import threading
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Sequence

from iac.utils.subproc import ControlledProcess

from .cel_eval import CelEngine
from .model import Book, Step
from .sensitive import SecretRedactor, sensitive_values

Status = Literal["success", "skipped", "failed", "failed_nonblocking"]


@dataclass(frozen=True)
class StepResult:
    book_path: str
    step: str
    status: Status
    exit_code: int
    duration_ms: int
    message: str = ""
    command: str = ""
    cwd: str = ""
    when: str = ""
    started_at: str = ""
    ended_at: str = ""
    stdin_file: str = ""
    stdout_file: str = ""
    stderr_file: str = ""
    attempts: int = 1
    max_attempts: int = 1
    outputs: Dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class RunOptions:
    quiet: bool = False
    stdin_file: Optional[str] = None
    stdout_file: Optional[str] = None
    stderr_file: Optional[str] = None
    file_mode: str = "append"
    artifacts_dir: Optional[str] = None
    capture: bool = False
    print_captured: bool = False
    pty: bool = False
    killpg: bool = True
    forward_signals: bool = True
    timeout_override: Optional[float] = None
    retries_override: Optional[int] = None
    retry_delay_override: Optional[float] = None
    verbose: bool = False
    run_id: str = ""


@dataclass
class _CaptureSinks:
    stdout_fh: Optional[object] = None
    stderr_fh: Optional[object] = None

    def close(self) -> None:
        for fh in (self.stdout_fh, self.stderr_fh):
            if fh is not None:
                try:
                    fh.close()
                except Exception:
                    pass


def _open_out_file(path: Optional[str], *, file_mode: str):
    if not path:
        return None
    mode = "ab" if file_mode == "append" else "wb"
    return open(path, mode)


class SignalForwarder:
    def __init__(
        self,
        proc: ControlledProcess,
        *,
        use_process_group: bool,
        forward_signals: Sequence[int],
    ) -> None:
        self.proc = proc
        self.use_process_group = use_process_group
        self.forward_signals = list(forward_signals)

        self._old_handlers: Dict[int, object] = {}
        self._received = threading.Event()
        self._last_signal: Optional[int] = None
        self._lock = threading.Lock()

    @property
    def received(self) -> bool:
        return self._received.is_set()

    @property
    def last_signal(self) -> Optional[int]:
        with self._lock:
            return self._last_signal

    def install(self) -> None:
        for sig in self.forward_signals:
            try:
                self._old_handlers[sig] = signal.getsignal(sig)
                signal.signal(sig, self._handler)  # type: ignore[arg-type]
            except Exception:
                continue

    def uninstall(self) -> None:
        for sig, old in self._old_handlers.items():
            try:
                signal.signal(sig, old)  # type: ignore[arg-type]
            except Exception:
                continue
        self._old_handlers.clear()

    def _handler(self, signum: int, _frame) -> None:
        with self._lock:
            self._last_signal = signum
        self._received.set()
        try:
            if self.use_process_group:
                self.proc.killpg(signum)
            else:
                self.proc.send_signal(signum)
        except Exception:
            pass


def _exit_code_for_signal(signum: int) -> int:
    return 128 + int(signum)


def _now_iso_utc() -> str:
    return dt.datetime.now(dt.timezone.utc).isoformat()


def _safe_name(name: str) -> str:
    return re.sub(r"[^A-Za-z0-9_.-]+", "_", name).strip("_") or "step"


def _resolve_file_setting(value: Any) -> Optional[str]:
    if value is None:
        return None
    s = str(value).strip()
    return s or None


def _ensure_parent(path: str) -> None:
    Path(path).expanduser().resolve().parent.mkdir(parents=True, exist_ok=True)


def _artifact_paths(
    opts: RunOptions, *, step_name: str, step_index: int
) -> tuple[Optional[str], Optional[str]]:
    if not opts.artifacts_dir:
        return None, None
    run_id = opts.run_id or "run"
    base = Path(opts.artifacts_dir).expanduser().resolve() / run_id
    base.mkdir(parents=True, exist_ok=True)
    stem = f"{step_index:03d}_{_safe_name(step_name)}"
    return (str(base / f"{stem}.stdout.log"), str(base / f"{stem}.stderr.log"))


def _execute_hook(
    hook_name: str,
    command: str,
    *,
    env: Dict[str, str],
    cwd: Optional[str],
    quiet: bool,
) -> int:
    proc = ControlledProcess(
        ["bash", "-lc", command],
        cwd=cwd,
        env=env,
        inherit_env=True,
        text=True,
        capture=not quiet,
        pty=False,
        merge_stderr=False,
        start_new_session=True,
    )
    proc.start()
    rc = proc.wait(timeout=None)
    return int(rc)


def _maybe_run_hook(
    book: Book,
    hook_name: str,
    *,
    book_path: Path,
    step_name: Optional[str],
    step_status: Optional[str],
    step_exit: Optional[int],
    run_id: str,
    report_path: Optional[str],
    quiet: bool,
) -> bool:
    hook = book.hooks.get(hook_name)
    if hook is None:
        return True

    env = {
        "IAC_HOOK_NAME": hook_name,
        "IAC_RUN_ID": run_id,
        "IAC_BOOK_PATH": str(book_path),
        "IAC_STEP_NAME": step_name or "",
        "IAC_STEP_STATUS": step_status or "",
        "IAC_STEP_EXIT_CODE": "" if step_exit is None else str(step_exit),
        "IAC_REPORT_PATH": report_path or "",
    }
    rc = _execute_hook(
        hook_name, hook.command, env=env, cwd=str(book_path.parent), quiet=quiet
    )
    if rc == 0:
        return True
    if hook.on_fail == "pass":
        return True
    if hook.on_fail == "warn":
        if not quiet:
            sys.stderr.write(
                f"WARNING: hook '{hook_name}' failed (warn), exit_code={rc}\n"
            )
            sys.stderr.flush()
        return True
    if not quiet:
        sys.stderr.write(f"hook '{hook_name}' failed (halt), exit_code={rc}\n")
        sys.stderr.flush()
    return False


def resolve_step(
    step: Step,
    *,
    book_path: Path,
    vars_map: Dict[str, Any],
    module_dirs: Optional[List[Path]] = None,
) -> tuple[
    List[str], str, Optional[str], Dict[str, str], Optional[float], Optional[str]
]:
    """Returns: argv, display, cwd, env, timeout, when_expr."""
    env: Dict[str, str] = {}
    cwd_s = step.cwd.strip() if isinstance(step.cwd, str) and step.cwd.strip() else None
    when_expr = step.when
    timeout = step.timeout

    if step.mode == "argv":
        if not step.argv:
            raise ValueError(f"step {step.name} is missing sh argv")
        argv = list(step.argv)
        return argv, shlex.join(argv), cwd_s, env, timeout, when_expr

    module_ref = (step.module or "").strip()
    if not module_ref:
        raise ValueError(f"step {step.name} is missing fn reference")

    env["IAC_MODULE_REF"] = module_ref
    env["IAC_MODULE_BOOK_PATH"] = str(book_path)
    env["IAC_MODULE_STEP_NAME"] = step.name
    env["IAC_MODULE_VARS_JSON"] = json.dumps(vars_map)
    if module_dirs:
        env["IAC_MODULE_DIRS"] = ":".join(str(d) for d in module_dirs)

    argv = [sys.executable, "-m", "iac.runbook.module_exec"]
    display = f"{sys.executable} -m iac.runbook.module_exec {module_ref}"
    return argv, display, cwd_s, env, timeout, when_expr


def run_one_step(
    book: Book,
    step: Step,
    *,
    book_path: Path,
    vars_map: Dict[str, Any],
    opts: RunOptions,
    step_index: int = 0,
    report_path: Optional[str] = None,
    module_dirs: Optional[List[Path]] = None,
    cel: Optional[CelEngine] = None,
    inputs_map: Optional[Dict[str, Any]] = None,
    steps_ctx: Optional[Dict[str, Any]] = None,
    run_ctx: Optional[Dict[str, Any]] = None,
) -> StepResult:
    cel_engine = cel if cel is not None else CelEngine.default()
    with_values: Dict[str, Any] = {}

    argv, display, cwd, env, timeout, when_expr = resolve_step(
        step,
        book_path=book_path,
        vars_map=vars_map,
        module_dirs=module_dirs,
    )

    if step.vars_map:
        eval_ctx = {
            "inputs": inputs_map or vars_map,
            "steps": steps_ctx or {},
            "run": run_ctx or {},
        }
        for key, expr in step.vars_map.items():
            with_values[key] = cel_engine.eval(expr, eval_ctx)
        env = dict(env)
        env.update({k: str(v) for k, v in with_values.items()})
        if step.mode == "module":
            module_vars = dict(vars_map)
            module_vars.update(with_values)
            env["IAC_MODULE_VARS_JSON"] = json.dumps(module_vars)

    file_mode = opts.file_mode if opts.file_mode in ("append", "truncate") else "append"

    art_stdout, art_stderr = _artifact_paths(
        opts, step_name=step.name, step_index=step_index
    )
    eff_stdin = _resolve_file_setting(opts.stdin_file)
    eff_stdout = _resolve_file_setting(opts.stdout_file) or art_stdout
    eff_stderr = _resolve_file_setting(opts.stderr_file) or art_stderr

    redactor = SecretRedactor.from_values(
        sensitive_values(book, vars_map={k: str(v) for k, v in vars_map.items()})
    )
    display_redacted = redactor.redact_text(display)

    if eff_stdout:
        _ensure_parent(eff_stdout)
    if eff_stderr:
        _ensure_parent(eff_stderr)
    if eff_stdin and not Path(eff_stdin).expanduser().exists():
        return StepResult(
            book_path=str(book_path),
            step=step.name,
            status="failed",
            exit_code=1,
            duration_ms=0,
            message=f"stdin file not found: {eff_stdin}",
            command=display_redacted,
            cwd=cwd or "",
            when=when_expr or "",
            started_at="",
            ended_at="",
            stdin_file=eff_stdin,
            stdout_file=eff_stdout or "",
            stderr_file=eff_stderr or "",
        )

    eff_timeout = (
        opts.timeout_override if opts.timeout_override is not None else timeout
    )
    eff_retries = (
        opts.retries_override if opts.retries_override is not None else step.retries
    )
    eff_retry_delay = (
        opts.retry_delay_override
        if opts.retry_delay_override is not None
        else step.retry_delay
    )
    max_attempts = 1 + max(0, int(eff_retries))
    eff_retry_delay = max(0.0, float(eff_retry_delay))

    if when_expr:
        ok = bool(
            cel_engine.eval(
                when_expr,
                {
                    "inputs": inputs_map or vars_map,
                    "steps": steps_ctx or {},
                    "run": run_ctx or {},
                    "vars": with_values,
                },
            )
        )
        if not ok:
            return StepResult(
                book_path=str(book_path),
                step=step.name,
                status="skipped",
                exit_code=0,
                duration_ms=0,
                message="when evaluated false",
                command=display_redacted,
                cwd=cwd or "",
                when=when_expr or "",
                started_at="",
                ended_at="",
                stdin_file=eff_stdin or "",
                stdout_file=eff_stdout or "",
                stderr_file=eff_stderr or "",
                attempts=1,
                max_attempts=max_attempts,
            )

    t0 = time.time()
    started_at = _now_iso_utc()

    if not _maybe_run_hook(
        book,
        "on_step_start",
        book_path=book_path,
        step_name=step.name,
        step_status=None,
        step_exit=None,
        run_id=opts.run_id,
        report_path=report_path,
        quiet=opts.quiet,
    ):
        ended_at = _now_iso_utc()
        return StepResult(
            book_path=str(book_path),
            step=step.name,
            status="failed",
            exit_code=1,
            duration_ms=0,
            message="on_step_start hook failed",
            command=display_redacted,
            cwd=cwd or "",
            when=when_expr or "",
            started_at=started_at,
            ended_at=ended_at,
            stdin_file=eff_stdin or "",
            stdout_file=eff_stdout or "",
            stderr_file=eff_stderr or "",
            attempts=1,
            max_attempts=max_attempts,
        )

    rc: int = 1
    msg = ""
    status: Status = "success"
    attempts_made = 0
    captured_proc: Optional[ControlledProcess] = None
    interrupted = False
    timed_out = False
    module_result_file: Optional[str] = None
    module_step_outputs: Dict[str, Any] = {}

    if step.mode == "module":
        fd, path = tempfile.mkstemp(
            prefix=f"iac_mod_{_safe_name(step.name)}_", suffix=".json"
        )
        os.close(fd)
        module_result_file = path
        env = dict(env)
        env["IAC_MODULE_RESULT_FILE"] = module_result_file
        env["IAC_MODULE_RUN_ID"] = opts.run_id

    for attempt in range(1, max_attempts + 1):
        attempts_made = attempt
        attempt_file_mode = file_mode if attempt == 1 else "append"
        sinks = _CaptureSinks(
            stdout_fh=_open_out_file(eff_stdout, file_mode=attempt_file_mode),
            stderr_fh=_open_out_file(eff_stderr, file_mode=attempt_file_mode),
        )

        def on_out_line(line: str) -> None:
            if opts.quiet:
                return
            sys.stdout.write(redactor.redact_text(line))
            sys.stdout.flush()

        def on_err_line(line: str) -> None:
            if opts.quiet:
                return
            sys.stderr.write(redactor.redact_text(line))
            sys.stderr.flush()

        def on_out_bytes(b: bytes) -> None:
            if sinks.stdout_fh is not None:
                try:
                    b = redactor.redact_bytes(b)
                    sinks.stdout_fh.write(b)
                    sinks.stdout_fh.flush()
                except Exception:
                    pass

        def on_err_bytes(b: bytes) -> None:
            if sinks.stderr_fh is not None:
                try:
                    b = redactor.redact_bytes(b)
                    sinks.stderr_fh.write(b)
                    sinks.stderr_fh.flush()
                except Exception:
                    pass

        p = ControlledProcess(
            argv,
            cwd=cwd,
            env=env or None,
            inherit_env=True,
            text=True,
            pty=opts.pty,
            capture=True,
            merge_stderr=False,
            start_new_session=True,
        )
        captured_proc = p

        sigs: List[int] = []
        if opts.forward_signals:
            sigs.extend([signal.SIGINT, signal.SIGTERM])
            if hasattr(signal, "SIGHUP"):
                sigs.append(signal.SIGHUP)
        forwarder = SignalForwarder(
            p, use_process_group=opts.killpg, forward_signals=sigs
        )

        try:
            p.start(
                on_stdout=on_out_line,
                on_stderr=on_err_line,
                on_stdout_bytes=on_out_bytes,
                on_stderr_bytes=on_err_bytes,
            )

            if eff_stdin:
                with open(eff_stdin, "rb") as fh:
                    p.write(fh.read())
                p.close_stdin()

            if opts.verbose:
                sys.stderr.write(f"[pid={p.pid} pgid={p.pgid}] {display_redacted}\n")
                sys.stderr.flush()

            forwarder.install()

            start_t = time.time()
            while True:
                polled = p.poll()
                if polled is not None:
                    rc = int(polled)
                    break

                if eff_timeout is not None and (time.time() - start_t) > eff_timeout:
                    sys.stderr.write(
                        f"Step timed out after {eff_timeout} seconds; stopping...\n"
                    )
                    sys.stderr.flush()
                    p.stop(grace=2.0, use_process_group=opts.killpg)
                    rc = 124
                    msg = "timeout"
                    timed_out = True
                    status = "failed"
                    break

                if forwarder.received:
                    time.sleep(0.2)
                    if p.is_running():
                        p.stop(grace=2.0, use_process_group=opts.killpg)
                    sig = forwarder.last_signal or signal.SIGINT
                    rc = _exit_code_for_signal(sig)
                    msg = f"terminated by signal {sig}"
                    interrupted = True
                    status = "failed"
                    break

                time.sleep(0.1)

        except FileNotFoundError as e:
            missing = getattr(e, "filename", None) or str(e)
            rc = 127
            msg = f"executable not found: {missing}"
            status = "failed"

        except KeyboardInterrupt:
            sys.stderr.write("Interrupted; stopping child process...\n")
            sys.stderr.flush()
            try:
                p.stop(grace=2.0, use_process_group=opts.killpg)
            finally:
                rc = 130
                msg = "keyboard interrupt"
                interrupted = True
                status = "failed"

        finally:
            forwarder.uninstall()
            sinks.close()

        if rc == 0:
            status = "success"
            break
        if interrupted:
            break
        if attempt < max_attempts:
            if not opts.quiet:
                sys.stderr.write(
                    f"retrying step '{step.name}' ({attempt + 1}/{max_attempts}) after exit_code={rc}\n"
                )
                sys.stderr.flush()
            if eff_retry_delay > 0:
                time.sleep(eff_retry_delay)

    dt_ms = int((time.time() - t0) * 1000)
    ended_at = _now_iso_utc()

    if rc != 0:
        if attempts_made > 1 and not msg:
            msg = f"failed after {attempts_made}/{max_attempts} attempts"
        elif timed_out and attempts_made > 1:
            msg = f"timeout after {attempts_made}/{max_attempts} attempts"
        if step.on_fail in ("pass", "warn"):
            status = "failed_nonblocking"
            if step.on_fail == "warn":
                sys.stderr.write(
                    f"WARNING: step '{step.name}' failed (warn), exit_code={rc}\n"
                )
                sys.stderr.flush()
        else:
            status = "failed"
    elif module_result_file:
        try:
            p = Path(module_result_file)
            if p.exists():
                data = json.loads(p.read_text(encoding="utf-8"))
                if isinstance(data, dict):
                    msg = str(data.get("message", "") or "")
                    raw_out = data.get("outputs")
                    if isinstance(raw_out, dict):
                        module_step_outputs = dict(raw_out)
        except Exception:
            pass
        finally:
            try:
                Path(module_result_file).unlink(missing_ok=True)
            except Exception:
                pass

    if (
        opts.capture
        and opts.print_captured
        and not opts.quiet
        and captured_proc is not None
    ):
        out = captured_proc.stdout_text()
        err = captured_proc.stderr_text()
        if out:
            sys.stdout.write(redactor.redact_text(out))
            sys.stdout.flush()
        if err:
            sys.stderr.write(redactor.redact_text(err))
            sys.stderr.flush()

    if not _maybe_run_hook(
        book,
        "on_step_end",
        book_path=book_path,
        step_name=step.name,
        step_status=status,
        step_exit=rc,
        run_id=opts.run_id,
        report_path=report_path,
        quiet=opts.quiet,
    ):
        status = "failed"
        rc = 1
        msg = "on_step_end hook failed"

    captured_stdout = ""
    captured_stderr = ""
    if captured_proc is not None:
        try:
            captured_stdout = captured_proc.stdout_text() or ""
            captured_stderr = captured_proc.stderr_text() or ""
        except Exception:
            captured_stdout = ""
            captured_stderr = ""

    step_outputs: Dict[str, Any] = {}
    if step.mode == "module" and module_step_outputs:
        step_outputs.update(module_step_outputs)
    if step.outputs_map:
        eval_ctx = {
            "inputs": inputs_map or vars_map,
            "steps": steps_ctx or {},
            "run": run_ctx or {},
            "vars": with_values,
            "result": {
                "stdout": captured_stdout,
                "stderr": captured_stderr,
                "exit_code": rc,
                "message": msg,
                "outputs": dict(step_outputs),
            },
        }
        for key, expr in step.outputs_map.items():
            step_outputs[key] = cel_engine.eval(expr, eval_ctx)

    return StepResult(
        book_path=str(book_path),
        step=step.name,
        status=status,
        exit_code=rc,
        duration_ms=dt_ms,
        message=msg,
        command=display_redacted,
        cwd=cwd or "",
        when=when_expr or "",
        started_at=started_at,
        ended_at=ended_at,
        stdin_file=eff_stdin or "",
        stdout_file=eff_stdout or "",
        stderr_file=eff_stderr or "",
        attempts=attempts_made,
        max_attempts=max_attempts,
        outputs=step_outputs,
    )


def run_book_sequential(
    book: Book,
    *,
    book_path: Path,
    vars_map: Dict[str, Any],
    opts: RunOptions,
    report_path: Optional[str] = None,
    module_dirs: Optional[List[Path]] = None,
) -> tuple[List[StepResult], Dict[str, Any]]:
    results: List[StepResult] = []
    run_outputs: Dict[str, Any] = {}
    cel = CelEngine.default()
    steps_ctx: Dict[str, Any] = {}
    run_ctx = {
        "id": opts.run_id,
        "book": str(book_path),
        "started_at": _now_iso_utc(),
    }
    if not _maybe_run_hook(
        book,
        "on_run_start",
        book_path=book_path,
        step_name=None,
        step_status=None,
        step_exit=None,
        run_id=opts.run_id,
        report_path=report_path,
        quiet=opts.quiet,
    ):
        return (
            [
                StepResult(
                    book_path=str(book_path),
                    step="__run_start__",
                    status="failed",
                    exit_code=1,
                    duration_ms=0,
                    message="on_run_start hook failed",
                )
            ],
            {},
        )

    for i, step in enumerate(book.steps):
        res = run_one_step(
            book,
            step,
            book_path=book_path,
            vars_map=vars_map,
            opts=opts,
            step_index=i,
            report_path=report_path,
            module_dirs=module_dirs,
            cel=cel,
            inputs_map=vars_map,
            steps_ctx=steps_ctx,
            run_ctx=run_ctx,
        )
        results.append(res)
        steps_ctx[step.name] = {
            "status": res.status,
            "succeeded": res.status in ("success", "failed_nonblocking"),
            "exit_code": res.exit_code,
            "message": res.message,
            "outputs": dict(res.outputs),
            "started_at": res.started_at,
            "ended_at": res.ended_at,
        }
        if res.status == "failed":
            break

    _maybe_run_hook(
        book,
        "on_run_end",
        book_path=book_path,
        step_name=(results[-1].step if results else None),
        step_status=(results[-1].status if results else None),
        step_exit=(results[-1].exit_code if results else None),
        run_id=opts.run_id,
        report_path=report_path,
        quiet=opts.quiet,
    )

    out_defs = book.outputs or {}
    eval_ctx = {
        "inputs": vars_map,
        "steps": steps_ctx,
        "run": run_ctx,
    }
    for name, spec in out_defs.items():
        when_expr = spec.get("when")
        if isinstance(when_expr, str) and when_expr.strip():
            try:
                ok = bool(cel.eval(when_expr, eval_ctx))
            except Exception:
                ok = False
            if not ok:
                continue
        value_expr = spec.get("value")
        if isinstance(value_expr, str) and value_expr.strip():
            try:
                run_outputs[name] = cel.eval(value_expr, eval_ctx)
            except Exception:
                run_outputs[name] = None

    return results, run_outputs



from __future__ import annotations

import datetime as dt
import json
import os
from pathlib import Path
import shlex
import signal
import sys
import threading
import time
from dataclasses import dataclass
from typing import Annotated, Dict, List, Optional, Sequence, Literal

import typer

from iac.cli.formatting import ensure_output_format, emit_structured, build_table
from iac.utils.subproc import ControlledProcess

app = typer.Typer(
    add_completion=False, no_args_is_help=True, help="Run one-off commands"
)

_SHELL_META = ("|", "&&", "||", ";", ">", "<", "$(", "`", "\n")


def _parse_env_kv(items: List[str]) -> Dict[str, str]:
    env: Dict[str, str] = {}
    for item in items:
        if "=" not in item:
            raise typer.BadParameter(
                f"Invalid --env value {item!r}. Expected KEY=VALUE."
            )
        k, v = item.split("=", 1)
        k = k.strip()
        if not k:
            raise typer.BadParameter(
                f"Invalid --env value {item!r}. KEY cannot be empty."
            )
        env[k] = v
    return env


def _default_shell_argv(cmd_str: str) -> List[str]:
    return ["bash", "-lc", cmd_str]


def _looks_like_shell_string(s: str) -> bool:
    return any(tok in s for tok in _SHELL_META)


def _ensure_parent(path: str) -> None:
    Path(path).expanduser().resolve().parent.mkdir(parents=True, exist_ok=True)


def _open_out_file(path: Optional[str], *, mode: str):
    if not path:
        return None
    _ensure_parent(path)
    file_mode = "ab" if mode == "append" else "wb"
    return open(path, file_mode)


def _load_stdin_bytes(
    *,
    stdin_file: Optional[str],
    stdin_text: Optional[str],
    stdin_json: Optional[str],
    stdin_env: Optional[str],
) -> Optional[bytes]:
    methods = [
        stdin_file is not None,
        stdin_text is not None,
        stdin_json is not None,
        stdin_env is not None,
    ]
    if sum(1 for m in methods if m) > 1:
        raise typer.BadParameter(
            "Use at most one of --stdin-file, --stdin-text, --stdin-json, or --stdin-env."
        )

    if stdin_file is not None:
        p = Path(stdin_file).expanduser()
        if not p.exists():
            raise typer.BadParameter(f"--stdin-file not found: {stdin_file}")
        return p.read_bytes()
    if stdin_text is not None:
        return stdin_text.encode("utf-8")
    if stdin_json is not None:
        try:
            obj = json.loads(stdin_json)
        except Exception as e:
            raise typer.BadParameter(f"Invalid --stdin-json payload: {e}") from e
        return json.dumps(obj).encode("utf-8")
    if stdin_env is not None:
        val = os.environ.get(stdin_env)
        if val is None:
            raise typer.BadParameter(f"--stdin-env variable is not set: {stdin_env}")
        return val.encode("utf-8")
    return None


@dataclass
class _CaptureSinks:
    stdout_fh: Optional[object] = None
    stderr_fh: Optional[object] = None

    def close(self) -> None:
        for fh in (self.stdout_fh, self.stderr_fh):
            if fh is not None:
                try:
                    fh.close()
                except Exception:
                    pass


class SignalForwarder:
    def __init__(
        self,
        proc: ControlledProcess,
        *,
        use_process_group: bool,
        forward_signals: Sequence[int],
    ) -> None:
        self.proc = proc
        self.use_process_group = use_process_group
        self.forward_signals = list(forward_signals)

        self._old_handlers: Dict[int, object] = {}
        self._received = threading.Event()
        self._last_signal: Optional[int] = None
        self._lock = threading.Lock()

    @property
    def received(self) -> bool:
        return self._received.is_set()

    @property
    def last_signal(self) -> Optional[int]:
        with self._lock:
            return self._last_signal

    def install(self) -> None:
        for sig in self.forward_signals:
            try:
                self._old_handlers[sig] = signal.getsignal(sig)
                signal.signal(sig, self._handler)  # type: ignore[arg-type]
            except Exception:
                continue

    def uninstall(self) -> None:
        for sig, old in self._old_handlers.items():
            try:
                signal.signal(sig, old)  # type: ignore[arg-type]
            except Exception:
                continue
        self._old_handlers.clear()

    def _handler(self, signum: int, _frame) -> None:
        with self._lock:
            self._last_signal = signum
        self._received.set()
        try:
            if self.use_process_group:
                self.proc.killpg(signum)
            else:
                self.proc.send_signal(signum)
        except Exception:
            pass


def _exit_code_for_signal(signum: int) -> int:
    return 128 + int(signum)


@app.command("cmd")
def cmd_(
    cmd: Annotated[
        str,
        typer.Argument(
            help="Command string. In --no-shell mode this is the executable."
        ),
    ],
    arg: Annotated[
        List[str],
        typer.Option(
            "--arg", "-a", help="Argument token for --no-shell mode (repeatable)"
        ),
    ] = [],
    shell: Annotated[
        bool,
        typer.Option(
            "--shell/--no-shell",
            help="Shell string (bash -lc) vs direct argv execution.",
        ),
    ] = True,
    cwd: Annotated[
        Optional[str], typer.Option("--cwd", help="Working directory")
    ] = None,
    env: Annotated[
        List[str],
        typer.Option("--env", help="Environment override(s), KEY=VALUE. Repeatable."),
    ] = [],
    stdin_file: Annotated[
        Optional[str], typer.Option("--stdin-file", help="Send stdin bytes from file")
    ] = None,
    stdin_text: Annotated[
        Optional[str], typer.Option("--stdin-text", help="Send UTF-8 text to stdin")
    ] = None,
    stdin_json: Annotated[
        Optional[str],
        typer.Option(
            "--stdin-json", help="Validate JSON then send compact JSON to stdin"
        ),
    ] = None,
    stdin_env: Annotated[
        Optional[str],
        typer.Option(
            "--stdin-env", help="Read stdin content from environment variable value"
        ),
    ] = None,
    clean_env: Annotated[
        bool, typer.Option("--clean-env", help="Do not inherit parent env")
    ] = False,
    pty: Annotated[bool, typer.Option("--pty", help="Run under a PTY")] = False,
    timeout: Annotated[
        Optional[float], typer.Option("--timeout", help="Timeout seconds")
    ] = None,
    killpg: Annotated[
        bool, typer.Option("--killpg/--no-killpg", help="Terminate process group")
    ] = True,
    forward_signals: Annotated[
        bool,
        typer.Option(
            "--forward-signals/--no-forward-signals",
            help="Forward SIGINT/SIGTERM(/SIGHUP)",
        ),
    ] = True,
    capture: Annotated[
        bool, typer.Option("--capture/--no-capture", help="Capture output in memory")
    ] = False,
    stdout_file: Annotated[
        Optional[str], typer.Option("--stdout-file", help="Append stdout bytes to file")
    ] = None,
    stderr_file: Annotated[
        Optional[str], typer.Option("--stderr-file", help="Append stderr bytes to file")
    ] = None,
    file_mode: Annotated[
        Literal["append", "truncate"],
        typer.Option(
            "--file-mode", help="File mode for stdout/stderr files: append|truncate"
        ),
    ] = "append",
    output: Annotated[
        str,
        typer.Option(
            "--output", help="Output format: table|json|yaml", show_default=True
        ),
    ] = "table",
    print_captured: Annotated[
        bool,
        typer.Option(
            "--print-captured/--no-print-captured",
            help="If --capture, print captured output at end",
        ),
    ] = False,
    quiet: Annotated[
        bool, typer.Option("-q", "--quiet", help="Do not stream output to console")
    ] = False,
    verbose: Annotated[
        bool, typer.Option("-v", "--verbose", help="Print extra metadata")
    ] = False,
) -> None:
    """Run a one-off command with shell/argv modes, stdin helpers, and output controls."""
    output_fmt = ensure_output_format(output)
    started_at = dt.datetime.now(dt.timezone.utc).isoformat()
    env_map = _parse_env_kv(env)
    stdin_bytes = _load_stdin_bytes(
        stdin_file=stdin_file,
        stdin_text=stdin_text,
        stdin_json=stdin_json,
        stdin_env=stdin_env,
    )

    if shell:
        if arg:
            raise typer.BadParameter(
                "In --shell mode, do not pass --arg values; include args in CMD."
            )
        proc_argv = _default_shell_argv(cmd)
        display = cmd
    else:
        proc_argv = [cmd] + arg
        display = shlex.join(proc_argv)
        if len(proc_argv) == 1 and _looks_like_shell_string(proc_argv[0]):
            raise typer.BadParameter(
                "It looks like you passed a shell expression while using --no-shell.\n"
                "Use: iac run cmd 'echo hi | wc -c'\n"
                "Or:  iac run cmd --no-shell echo --arg hi\n"
            )

    sinks = _CaptureSinks(
        stdout_fh=_open_out_file(stdout_file, mode=file_mode),
        stderr_fh=_open_out_file(stderr_file, mode=file_mode),
    )

    def on_out_line(line: str) -> None:
        if quiet:
            return
        sys.stdout.write(line)
        sys.stdout.flush()

    def on_err_line(line: str) -> None:
        if quiet:
            return
        sys.stderr.write(line)
        sys.stderr.flush()

    def on_out_bytes(b: bytes) -> None:
        if sinks.stdout_fh is not None:
            try:
                sinks.stdout_fh.write(b)
                sinks.stdout_fh.flush()
            except Exception:
                pass

    def on_err_bytes(b: bytes) -> None:
        if sinks.stderr_fh is not None:
            try:
                sinks.stderr_fh.write(b)
                sinks.stderr_fh.flush()
            except Exception:
                pass

    p = ControlledProcess(
        proc_argv,
        cwd=cwd,
        env=env_map if env_map else None,
        inherit_env=not clean_env,
        text=True,
        pty=pty,
        capture=capture,
        merge_stderr=False,
        start_new_session=True,
    )

    sigs: List[int] = []
    if forward_signals:
        sigs.extend([signal.SIGINT, signal.SIGTERM])
        if hasattr(signal, "SIGHUP"):
            sigs.append(signal.SIGHUP)

    forwarder = SignalForwarder(p, use_process_group=killpg, forward_signals=sigs)

    start_t = time.time()
    exit_code = 1
    message = ""

    try:
        p.start(
            on_stdout=on_out_line,
            on_stderr=on_err_line,
            on_stdout_bytes=on_out_bytes,
            on_stderr_bytes=on_err_bytes,
        )
        if stdin_bytes is not None:
            p.write(stdin_bytes)
            p.close_stdin()
        if verbose:
            typer.echo(f"[pid={p.pid} pgid={p.pgid}] {display}", err=True)
        forwarder.install()

        while True:
            rc = p.poll()
            if rc is not None:
                exit_code = int(rc)
                break
            if timeout is not None and (time.time() - start_t) > timeout:
                typer.echo(
                    f"Command timed out after {timeout} seconds; stopping...", err=True
                )
                p.stop(grace=2.0, use_process_group=killpg)
                exit_code = 124
                message = "timeout"
                break
            if forwarder.received:
                time.sleep(0.2)
                if p.is_running():
                    p.stop(grace=2.0, use_process_group=killpg)
                sig = forwarder.last_signal or signal.SIGINT
                exit_code = _exit_code_for_signal(sig)
                message = f"terminated by signal {sig}"
                break
            time.sleep(0.1)

    except FileNotFoundError as e:
        missing = getattr(e, "filename", None) or str(e)
        typer.echo(f"Executable not found: {missing}", err=True)
        exit_code = 127
        message = f"executable not found: {missing}"

    except KeyboardInterrupt:
        typer.echo("Interrupted; stopping child process...", err=True)
        try:
            p.stop(grace=2.0, use_process_group=killpg)
        finally:
            exit_code = 130
            message = "keyboard interrupt"

    finally:
        forwarder.uninstall()
        sinks.close()

    ended_at = dt.datetime.now(dt.timezone.utc).isoformat()
    rc_final = (
        exit_code
        if exit_code is not None
        else (p.returncode if p.returncode is not None else 1)
    )
    duration_ms = int((time.time() - start_t) * 1000)

    if capture and print_captured and not quiet:
        out = p.stdout_text()
        err = p.stderr_text()
        if out:
            sys.stdout.write(out)
            sys.stdout.flush()
        if err:
            sys.stderr.write(err)
            sys.stderr.flush()

    payload = {
        "kind": "cmd",
        "started_at": started_at,
        "ended_at": ended_at,
        "duration_ms": duration_ms,
        "exit_code": int(rc_final),
        "status": "success" if int(rc_final) == 0 else "failed",
        "message": message,
        "shell": shell,
        "command": display,
        "argv": proc_argv,
        "cwd": cwd or "",
        "stdin_source": (
            "file"
            if stdin_file is not None
            else (
                "text"
                if stdin_text is not None
                else (
                    "json"
                    if stdin_json is not None
                    else "env" if stdin_env is not None else ""
                )
            )
        ),
        "stdin_bytes": len(stdin_bytes) if stdin_bytes is not None else 0,
        "stdout_file": stdout_file or "",
        "stderr_file": stderr_file or "",
        "file_mode": file_mode,
    }

    if output_fmt in ("json", "yaml"):
        emit_structured(payload, output_fmt)
        raise typer.Exit(code=int(rc_final))

    table = build_table(
        [
            "STATUS",
            "EXIT",
            "DURATION_MS",
            "COMMAND",
            "CWD",
            "STDIN",
            "STDOUT_FILE",
            "STDERR_FILE",
            "MESSAGE",
        ],
        [
            [
                payload["status"],
                payload["exit_code"],
                payload["duration_ms"],
                payload["command"],
                payload["cwd"],
                payload["stdin_source"],
                payload["stdout_file"],
                payload["stderr_file"],
                payload["message"],
            ]
        ],
    )
    typer.echo("Run Cmd")
    typer.echo(table.get_string())
    raise typer.Exit(code=int(rc_final))



from __future__ import annotations

from pathlib import Path

import pytest

from iac.cli.formatting import build_table, ensure_output_format
from iac.runbook.cel_eval import CelEngine
from iac.runbook.deps import resolve_install_command
from iac.runbook.discovery import (
    book_id_for_path,
    list_books,
    parse_book_dirs,
    resolve_book,
    split_book_step,
)
from iac.runbook.loader import load_book, parse_vars_kv, resolve_inputs
from iac.runbook.model import parse_book
from iac.runbook.sensitive import (
    REDACTION,
    SecretRedactor,
    redact_vars_map,
    sensitive_values,
)


def test_formatting_helpers() -> None:
    assert ensure_output_format("json") == "json"
    with pytest.raises(Exception):
        ensure_output_format("nope")
    t = build_table(["A", "B"], [[None, True], ["x", False]])
    s = t.get_string()
    assert "true" in s and "false" in s and "x" in s


def test_discovery_and_resolution(tmp_path: Path) -> None:
    d1 = tmp_path / "books"
    d2 = tmp_path / "other"
    d1.mkdir()
    d2.mkdir()
    (d1 / "a.yaml").write_text(
        "kind: pyplines.runbook.v1\nname: a\nsteps: [{name: s, sh: ['echo','hi']}]\n",
        encoding="utf-8",
    )
    (d2 / "b").mkdir()
    (d2 / "b" / "book.yaml").write_text(
        "kind: pyplines.runbook.v1\nname: b\nsteps: [{name: s, sh: ['echo','hi']}]\n",
        encoding="utf-8",
    )

    dirs = parse_book_dirs(f"{d1}:{d2}:{d1}")
    books = list_books(dirs)
    assert len(books) >= 2
    assert book_id_for_path(d2 / "b" / "book.yaml") == "b"
    assert resolve_book("a", [d1]).name == "a.yaml"
    assert resolve_book("b", [d2]).name == "book.yaml"
    assert split_book_step("a@x") == ("a", "x")
    with pytest.raises(ValueError):
        split_book_step("broken")


def test_loader_and_parse_helpers(tmp_path: Path) -> None:
    assert parse_vars_kv(["a=1"]) == {"a": "1"}
    with pytest.raises(ValueError):
        parse_vars_kv(["bad"])

    p = tmp_path / "book.yaml"
    p.write_text(
        "kind: pyplines.runbook.v1\nname: b\nsteps: [{name: s, sh: ['echo','hi']}]\n",
        encoding="utf-8",
    )
    b = load_book(p)
    assert b.meta.name == "b"


def test_model_validation_and_sensitive() -> None:
    data = {
        "kind": "pyplines.runbook.v1",
        "name": "x",
        "inputs": {"secret": {"type": "string", "default": "abc"}},
        "steps": [
            {
                "name": "s",
                "sh": ["bash", "-lc", "echo hi"],
                "retries": 1,
                "retry_delay": 0.1,
            }
        ],
    }
    b = parse_book(data)
    vals = sensitive_values(b, vars_map={"secret": "abc"})
    red = SecretRedactor.from_values(vals)
    assert red.redact_text("abc") == "abc"
    assert redact_vars_map(b, {"secret": "abc"})["secret"] == "abc"

    with pytest.raises(ValueError):
        parse_book({"version": 1, "steps": [{"name": "s", "sh": ["echo", "x"]}]})
    with pytest.raises(ValueError):
        parse_book(
            {
                "kind": "pyplines.runbook.v1",
                "steps": [{"name": "s", "sh": ["echo", "x"], "retries": -1}],
            }
        )
    with pytest.raises(ValueError):
        parse_book(
            {
                "kind": "pyplines.runbook.v1",
                "steps": [{"name": "s", "sh": ["echo", "x"], "retry_delay": -1}],
            }
        )
    with pytest.raises(ValueError):
        parse_book(
            {
                "kind": "pyplines.runbook.v1",
                "steps": [{"name": "s", "sh": ["echo", "x"], "timeout": 0}],
            }
        )


def test_resolve_install_command() -> None:
    target, cmd, tried = resolve_install_command({"darwin": "brew i"})
    assert isinstance(tried, list)
    if target:
        assert cmd


def test_packages_parse_defaults() -> None:
    b = parse_book(
        {
            "kind": "pyplines.runbook.v1",
            "name": "x",
            "packages": {
                "os": [
                    {
                        "name": "bash",
                        "min_version": ">=3.2.0",
                        "check": "bash --version",
                    }
                ],
                "py": [{"name": "requests", "min_version": ">=2.32.5"}],
            },
            "steps": [{"name": "s", "sh": ["echo", "x"]}],
        }
    )
    assert len(b.requires) == 2
    bash = next(r for r in b.requires if r.name == "bash")
    reqs = next(r for r in b.requires if r.name == "requests")
    assert bash.req_type == "os"
    assert bash.install_cmd == "dnf install -y bash"
    assert reqs.req_type == "py"
    assert reqs.install_cmd == "python3 -m pip install requests>=2.32.5"


def test_kind_v1_parse_and_input_resolution_and_cel() -> None:
    book = parse_book(
        {
            "kind": "pyplines.runbook.v1",
            "name": "billing",
            "inputs": {
                "month": {"type": "string", "regex": r"^\d{4}-\d{2}$"},
                "accounts": {"type": "array", "default": ["prod"], "minItems": 1},
                "include_pdf": {"type": "boolean", "default": False},
            },
            "packages": {
                "os": [{"name": "bash", "check": "bash --version"}],
                "py": [{"name": "requests", "min_version": ">=2.32.5"}],
            },
            "steps": [
                {
                    "name": "export",
                    "sh": ["python3", "-c", "print('ok')"],
                    "vars": {"MONTH": "inputs.month"},
                    "outputs": {"status": '"ok"'},
                }
            ],
            "outputs": {"report_url": {"type": "string", "value": '"https://x"'}},
        }
    )
    assert book.meta.kind == "pyplines.runbook.v1"
    assert book.steps[0].mode == "argv"
    assert book.steps[0].vars_map["MONTH"] == "inputs.month"
    assert book.outputs["report_url"]["type"] == "string"

    vals = resolve_inputs(
        book.inputs,
        input_cli=["month=2026-03", 'accounts=["prod","dev"]', "include_pdf=true"],
    )
    assert vals["month"] == "2026-03"
    assert vals["accounts"] == ["prod", "dev"]
    assert vals["include_pdf"] is True

    cel = CelEngine.default()
    got = cel.eval(
        "size(inputs.accounts) > 0", {"inputs": vals, "steps": {}, "run": {}}
    )
    assert bool(got) is True
    parsed = cel.eval(
        'json("{\\"x\\":1}").x',
        {"inputs": vals, "steps": {}, "run": {}},
    )
    assert parsed == 1
    s = cel.eval("to_json(inputs.accounts)", {"inputs": vals, "steps": {}, "run": {}})
    assert "prod" in s


def test_resolve_inputs_error_paths() -> None:
    specs = {
        "month": {"type": "string", "regex": r"^\d{4}-(0[1-9]|1[0-2])$"},
        "flag": {"type": "boolean"},
        "items": {"type": "array", "minItems": 1},
        "n": {"type": "integer"},
        "f": {"type": "number"},
    }
    with pytest.raises(ValueError):
        resolve_inputs(
            specs, input_cli=["month=bad", "flag=true", "items=a", "n=1", "f=1.2"]
        )
    with pytest.raises(ValueError):
        resolve_inputs(
            specs, input_cli=["month=2026-03", "flag=maybe", "items=a", "n=1", "f=1.2"]
        )
    with pytest.raises(ValueError):
        resolve_inputs(
            specs, input_cli=["month=2026-03", "flag=true", "items=", "n=1", "f=1.2"]
        )
    with pytest.raises(ValueError):
        resolve_inputs(
            specs, input_cli=["month=2026-03", "flag=true", "items=a", "n=x", "f=1.2"]
        )
    with pytest.raises(ValueError):
        resolve_inputs(
            specs, input_cli=["month=2026-03", "flag=true", "items=a", "n=1", "f=x"]
        )
